// Copyright 2024 The Bazel Authors. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

/** Verify that sun.jnu.encoding is UTF-8. */
public class CheckSunJnuEncoding {
  public static void main(String[] args) {
    String sunJnuEncoding = System.getProperty("sun.jnu.encoding");
    if (!"UTF-8".equals(sunJnuEncoding)) {
      System.err.println("ERROR: sun.jnu.encoding is not UTF-8: " + sunJnuEncoding);
      System.exit(1);
    }
  }
}
