// Copyright 2024 The Bazel Authors. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

syntax = "proto3";

package crash_debugging;

option java_package = "com.google.devtools.build.lib.runtime";
option java_outer_classname = "CrashDebuggingProtos";

// Structured data to assist with debugging crashes (including OOMs).
message CrashDebugging {
  // Information about actions that were executing when the crash occurred.
  optional InflightActionInfo inflight_actions = 1;
}

message InflightActionInfo {
  // Total number of actions executing.
  optional int32 count = 1;

  message InflightAction {
    // The action mnemonic.
    optional string mnemonic = 1;
    // Number of actions with this mnemonic that were executing.
    optional int32 count = 2;
  }

  // Top 20 actions executing by mnemonic.
  repeated InflightAction top_mnemonics = 2;
}
