// Copyright 2025 The Bazel Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package com.google.testing.coverage;

/** An expression that always evaluates to false. */
public final class NullExpression implements CoverageExpression {

  public static final NullExpression NULL_EXP = new NullExpression();

  private NullExpression() {}

  @Override
  public boolean eval(final boolean[] probes) {
    return false;
  }
}
