// Copyright 2025 The Bazel Authors. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package com.google.devtools.build.lib.util;

import org.commonmark.parser.Parser;
import org.commonmark.renderer.html.HtmlRenderer;

/** Helpers for markdown. */
public class Markdown {
  private static final Parser PARSER = Parser.builder().build();
  private static final HtmlRenderer HTML_RENDERER = HtmlRenderer.builder().escapeHtml(true).build();

  /** Parses a string as markdown and renders it as HTML. */
  public static String renderToHtml(String md) {
    return HTML_RENDERER.render(PARSER.parse(md)).trim();
  }
}
