<p>Boolean; <a href="#configurable-attributes">nonconfigurable</a>;
  default is <code>False</code></p>

<p>Forces the test to be run locally, without sandboxing.</p>

<p>Setting this to True is equivalent to providing "local" as a tag
(<code>tags=["local"]</code>).</p>
