<p>Non-negative integer less than or equal to 50; default is <code>-1</code></p>

<p>Specifies the number of parallel shards
to use to run the test.</p>

<p>If set, this value will override any heuristics used to determine the number of
parallel shards with which to run the test. Note that for some test
rules, this parameter may be required to enable sharding
in the first place. Also see <code>--test_sharding_strategy</code>.</p>

<p>If test sharding is enabled, the environment variable <code>
<a href="${link test-encyclopedia#initial-conditions}">TEST_TOTAL_SHARDS</a>
</code> will be set to this value when spawning the test.</p>

<p>Sharding requires the test runner to support the test sharding protocol.
If it does not, then it will most likely run every test in every shard, which
is not what you want.</p>

<p>See
<a href="${link test-encyclopedia#test-sharding}">Test Sharding</a>
in the Test Encyclopedia for details on sharding.</p>
