// Copyright 2026 The Bazel Authors. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package com.google.devtools.build.lib.remote.chunking;

import build.bazel.remote.execution.v2.Digest;
import com.google.common.collect.ImmutableList;
import java.io.IOException;
import java.io.InputStream;

/** Splits a blob into content-defined chunks. */
public interface ContentDefinedChunker {

  /**
   * Chunks a blob and returns the chunk digests in order.
   *
   * <p>This method is used for building MerkleTree entries for large files and for chunked uploads.
   * It only returns the content digests; callers that need the raw chunk data can read it from the
   * original (seekable) file using the chunk sizes, similar to how whole blobs work.
   *
   * <p>Chunking is deterministic: the same input always produces the same chunks for a given
   * configuration.
   *
   * @throws IOException only if reading from {@code input} throws; chunking and digest computation
   *     themselves do not perform any I/O
   */
  ImmutableList<Digest> chunkToDigests(InputStream input) throws IOException;
}
