// Copyright 2026 The Bazel Authors. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

edition = "2024";

package execution_graph_writer;

option java_package = "com.google.devtools.build.lib.runtime";

// Statistics about the execution graph writer in ExecutionGraphModule.
message ExecutionGraphWriterStats {
  // Number of milliseconds threads were blocked waiting for the queue to have
  // space (sum over all threads).
  int64 blocked_millis = 1;
  // Peak number of bytes queued for writing.
  int64 max_queued_bytes = 2;
}
