// Copyright 2026 The Bazel Authors. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package com.google.devtools.build.skyframe;

/** Factory for {@link NodeVersion} instances. */
public final class NodeVersionGetter {

  private NodeVersionGetter() {}

  public static NodeVersion get(Version lastChanged, Version lastEvaluated) {
    if (lastChanged.equals(lastEvaluated)) {
      return lastChanged;
    }
    return new ChangePruned(lastChanged, lastEvaluated);
  }

  /**
   * Basic implementation of {@link NodeVersion} for the case where {@link #lastChanged} and {@link
   * #lastEvaluated} are different versions.
   */
  private record ChangePruned(Version lastChanged, Version lastEvaluated) implements NodeVersion {}
}
